/*
 * ACPI PCI HotPlug PCI configuration space management
 *
 * Copyright (C) 1995,2001 Compaq Computer Corporation
 * Copyright (C) 2001 Greg Kroah-Hartman (greg@kroah.com)
 * Copyright (C) 2001,2002 IBM Corp.
 * Copyright (C) 2002 Takayoshi Kochi (t-kochi@bq.jp.nec.com)
 * Copyright (C) 2002 Hiroshi Aono (h-aono@ap.jp.nec.com)
 * Copyright (C) 2002 NEC Corporation
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, GOOD TITLE or
 * NON INFRINGEMENT.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Send feedback to <t-kochi@bq.jp.nec.com>
 *
 */

#include <linux/init.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/acpi.h>
#include "../pci.h"
#include "pci_hotplug.h"
#include "acpiphp.h"

#define MY_NAME "acpiphp_pci"

/* P2P bridge hotplug support */
static int init_config_space_behind_bridge (struct acpiphp_func *func);
static struct pci_resource *find_secondary_busno (struct acpiphp_func *func);
static int scan_behind_bridge (struct acpiphp_func *func);
static int init_base_address_registers (struct acpiphp_func *func, u8 devno, u8 funcno, struct acpiphp_func *usable_res, struct acpiphp_func *used_res);

/* allocate mem/pmem/io resource to a new function */
static int init_config_space (struct acpiphp_func *func)
{
	u32 bar, len;
	u32 address[] = {
		PCI_BASE_ADDRESS_0,
		PCI_BASE_ADDRESS_1,
		PCI_BASE_ADDRESS_2,
		PCI_BASE_ADDRESS_3,
		PCI_BASE_ADDRESS_4,
		PCI_BASE_ADDRESS_5,
		0
	};
	int count;
	struct acpiphp_bridge *bridge;
	struct pci_resource *res;
	struct pci_bus *pbus;
	int bus, device, function;
	unsigned int devfn;
	u16 tmp;
	u8 hdrtype;		/* P2P bridge hotplug support */

	bridge = func->slot->bridge;
	pbus = bridge->pci_bus;
	bus = bridge->bus;
	device = func->slot->device;
	function = func->function;
	devfn = PCI_DEVFN(device, function);

	/* check header type for target device. */
	if (acpiphp_hotplug_bridge) 
		pci_bus_read_config_byte (pbus, devfn, PCI_HEADER_TYPE, &hdrtype);

	for (count = 0; address[count]; count++) {	/* for 6 BARs */

		/* check 0x10,0x14 base address register for bridge device. */
		if (acpiphp_hotplug_bridge) {
			if ((hdrtype & PCI_HEADER_TYPE_BRIDGE) && (count > 1)) {
				break;
			}
		}

		pci_bus_write_config_dword(pbus, devfn,
					   address[count], 0xFFFFFFFF);
		pci_bus_read_config_dword(pbus, devfn, address[count], &bar);

		if (!bar)	/* This BAR is not implemented */
			continue;

		dbg("Device %02x.%02x BAR %d wants %x\n", device, function, count, bar);

		if (bar & PCI_BASE_ADDRESS_SPACE_IO) {
			/* This is IO */

			len = bar & (PCI_BASE_ADDRESS_IO_MASK & 0xFFFF);
			len = len & ~(len - 1);

			dbg("len in IO %x, BAR %d\n", len, count);

			spin_lock(&bridge->res_lock);
			res = acpiphp_get_io_resource(&bridge->io_head, len);
			spin_unlock(&bridge->res_lock);

			if (!res) {
				err("cannot allocate requested io for %02x:%02x.%d len %x\n",
				    bus, device, function, len);
				return -1;
			}
			pci_bus_write_config_dword(pbus, devfn,
						   address[count],
						   (u32)res->base);
			res->next = func->io_head;
			func->io_head = res;

		} else {
			/* This is Memory */
			if (bar & PCI_BASE_ADDRESS_MEM_PREFETCH) {
				/* pfmem */

				len = bar & 0xFFFFFFF0;
				len = ~len + 1;

				dbg("len in PFMEM %x, BAR %d\n", len, count);

				spin_lock(&bridge->res_lock);
				res = acpiphp_get_resource(&bridge->p_mem_head, len);
				spin_unlock(&bridge->res_lock);

				if (!res) {
					err("cannot allocate requested pfmem for %02x:%02x.%d len %x\n",
					    bus, device, function, len);
					return -1;
				}

				pci_bus_write_config_dword(pbus, devfn,
							   address[count],
							   (u32)res->base);

				if (bar & PCI_BASE_ADDRESS_MEM_TYPE_64) {	/* takes up another dword */
					dbg("inside the pfmem 64 case, count %d\n", count);
					count += 1;
					pci_bus_write_config_dword(pbus, devfn,
								   address[count],
								   (u32)(res->base >> 32));
				}

				res->next = func->p_mem_head;
				func->p_mem_head = res;

			} else {
				/* regular memory */

				len = bar & 0xFFFFFFF0;
				len = ~len + 1;

				dbg("len in MEM %x, BAR %d\n", len, count);

				spin_lock(&bridge->res_lock);
				res = acpiphp_get_resource(&bridge->mem_head, len);
				spin_unlock(&bridge->res_lock);

				if (!res) {
					err("cannot allocate requested pfmem for %02x:%02x.%d len %x\n",
					    bus, device, function, len);
					return -1;
				}

				pci_bus_write_config_dword(pbus, devfn,
							   address[count],
							   (u32)res->base);

				if (bar & PCI_BASE_ADDRESS_MEM_TYPE_64) {
					/* takes up another dword */
					dbg("inside mem 64 case, reg. mem, count %d\n", count);
					count += 1;
					pci_bus_write_config_dword(pbus, devfn,
								   address[count],
								   (u32)(res->base >> 32));
				}

				res->next = func->mem_head;
				func->mem_head = res;

			}
		}
	}

	/* disable expansion rom */
	pci_bus_write_config_dword(pbus, devfn, PCI_ROM_ADDRESS, 0x00000000);

	if (acpiphp_hotplug_bridge) {
		/* if device is bridge, check and initialize behind bridge device. */
		if (hdrtype & PCI_HEADER_TYPE_BRIDGE) {
			int ret = 0;

			/* check and ini behind bridge device. */
			ret = init_config_space_behind_bridge (func);
			if (ret < 0) {
				return -1;
			}
		}
		else {
			/* set PCI parameters from _HPP */
			pci_bus_write_config_byte(pbus, devfn, PCI_CACHE_LINE_SIZE,
						  bridge->hpp.cache_line_size);
			pci_bus_write_config_byte(pbus, devfn, PCI_LATENCY_TIMER,
						  bridge->hpp.latency_timer);

			pci_bus_read_config_word(pbus, devfn, PCI_COMMAND, &tmp);
			if (bridge->hpp.enable_SERR)
				tmp |= PCI_COMMAND_SERR;
			if (bridge->hpp.enable_PERR)
				tmp |= PCI_COMMAND_PARITY;
			pci_bus_write_config_word(pbus, devfn, PCI_COMMAND, tmp);
		}
	}
	else {
		/* set PCI parameters from _HPP */
		pci_bus_write_config_byte(pbus, devfn, PCI_CACHE_LINE_SIZE,
					  bridge->hpp.cache_line_size);
		pci_bus_write_config_byte(pbus, devfn, PCI_LATENCY_TIMER,
					  bridge->hpp.latency_timer);

		pci_bus_read_config_word(pbus, devfn, PCI_COMMAND, &tmp);
		if (bridge->hpp.enable_SERR)
			tmp |= PCI_COMMAND_SERR;
		if (bridge->hpp.enable_PERR)
			tmp |= PCI_COMMAND_PARITY;
		pci_bus_write_config_word(pbus, devfn, PCI_COMMAND, tmp);
	}

	return 0;
}

/* detect_used_resource - subtract resource under dev from bridge */
static int detect_used_resource (struct acpiphp_bridge *bridge, struct pci_dev *dev)
{
	int count;

	dbg("Device %s\n", pci_name(dev));

	for (count = 0; count < DEVICE_COUNT_RESOURCE; count++) {
		struct pci_resource *res;
		struct pci_resource **head;
		unsigned long base = dev->resource[count].start;
		unsigned long len = dev->resource[count].end - base + 1;
		unsigned long flags = dev->resource[count].flags;

		if (!flags)
			continue;

		dbg("BAR[%d] 0x%lx - 0x%lx (0x%lx)\n", count, base,
				base + len - 1, flags);

		if (flags & IORESOURCE_IO) {
			head = &bridge->io_head;
		} else if (flags & IORESOURCE_PREFETCH) {
			head = &bridge->p_mem_head;
		} else {
			head = &bridge->mem_head;
		}

		spin_lock(&bridge->res_lock);
		res = acpiphp_get_resource_with_base(head, base, len);
		spin_unlock(&bridge->res_lock);
		if (res)
			kfree(res);
	}

	/* detect used bus resource */
	if (acpiphp_hotplug_bridge) {
		if (dev->hdr_type & PCI_HEADER_TYPE_BRIDGE) {
			for (count = dev->bus->primary; count <= dev->bus->subordinate; count++) {
				struct pci_resource *res;

				spin_lock (&bridge->res_lock);
				res = acpiphp_get_resource_with_base (&bridge->bus_head, (u64)count, 1);
				spin_unlock (&bridge->res_lock);
				if (res) {
					kfree (res);
				}
			}
		}
	}

	return 0;
}


/**
 * acpiphp_detect_pci_resource - detect resources under bridge
 * @bridge: detect all resources already used under this bridge
 *
 * collect all resources already allocated for all devices under a bridge.
 */
int acpiphp_detect_pci_resource (struct acpiphp_bridge *bridge)
{
	struct list_head *l;
	struct pci_dev *dev;

	list_for_each (l, &bridge->pci_bus->devices) {
		dev = pci_dev_b(l);
		detect_used_resource(bridge, dev);
	}

	return 0;
}


/**
 * acpiphp_init_slot_resource - gather resource usage information of a slot
 * @slot: ACPI slot object to be checked, should have valid pci_dev member
 *
 * TBD: PCI-to-PCI bridge case
 *      use pci_dev->resource[]
 */
int acpiphp_init_func_resource (struct acpiphp_func *func)
{
	u64 base;
	u32 bar, len;
	u32 address[] = {
		PCI_BASE_ADDRESS_0,
		PCI_BASE_ADDRESS_1,
		PCI_BASE_ADDRESS_2,
		PCI_BASE_ADDRESS_3,
		PCI_BASE_ADDRESS_4,
		PCI_BASE_ADDRESS_5,
		0
	};
	int count;
	struct pci_resource *res;
	struct pci_dev *dev;

	dev = func->pci_dev;
	dbg("Hot-pluggable device %s\n", pci_name(dev));

	for (count = 0; address[count]; count++) {	/* for 6 BARs */

		/* check 0x10,0x14 base address register for bridge device. */
		if (acpiphp_hotplug_bridge) {
			if ((dev->hdr_type & PCI_HEADER_TYPE_BRIDGE) && (count > 1)) {
				break;
			}
		}

		pci_read_config_dword(dev, address[count], &bar);

		if (!bar)	/* This BAR is not implemented */
			continue;

		pci_write_config_dword(dev, address[count], 0xFFFFFFFF);
		pci_read_config_dword(dev, address[count], &len);

		if (len & PCI_BASE_ADDRESS_SPACE_IO) {
			/* This is IO */
			base = bar & 0xFFFFFFFC;
			len = len & (PCI_BASE_ADDRESS_IO_MASK & 0xFFFF);
			len = len & ~(len - 1);

			dbg("BAR[%d] %08x - %08x (IO)\n", count, (u32)base, (u32)base + len - 1);

			res = acpiphp_make_resource(base, len);
			if (!res)
				goto no_memory;

			res->next = func->io_head;
			func->io_head = res;

		} else {
			/* This is Memory */
			base = bar & 0xFFFFFFF0;
			if (len & PCI_BASE_ADDRESS_MEM_PREFETCH) {
				/* pfmem */

				len &= 0xFFFFFFF0;
				len = ~len + 1;

				if (len & PCI_BASE_ADDRESS_MEM_TYPE_64) {	/* takes up another dword */
					dbg("prefetch mem 64\n");
					count += 1;
				}
				dbg("BAR[%d] %08x - %08x (PMEM)\n", count, (u32)base, (u32)base + len - 1);
				res = acpiphp_make_resource(base, len);
				if (!res)
					goto no_memory;

				res->next = func->p_mem_head;
				func->p_mem_head = res;

			} else {
				/* regular memory */

				len &= 0xFFFFFFF0;
				len = ~len + 1;

				if (len & PCI_BASE_ADDRESS_MEM_TYPE_64) {
					/* takes up another dword */
					dbg("mem 64\n");
					count += 1;
				}
				dbg("BAR[%d] %08x - %08x (MEM)\n", count, (u32)base, (u32)base + len - 1);
				res = acpiphp_make_resource(base, len);
				if (!res)
					goto no_memory;

				res->next = func->mem_head;
				func->mem_head = res;

			}
		}

		pci_write_config_dword(dev, address[count], bar);
	}

	/* if device is bridge, set base and limit resources from dev->resource[] */
	if (acpiphp_hotplug_bridge) {
		if (dev->hdr_type & PCI_HEADER_TYPE_BRIDGE) {
			int i;
			struct resource *r;
			struct pci_bus *bus;

			/* check bridge resource range */
			for (i = PCI_BRIDGE_RESOURCES; i < PCI_NUM_RESOURCES; i++) {
				res = NULL;
				r = &dev->resource[i];

				/* I/O */
				if (r->flags & IORESOURCE_IO) {
					res = acpiphp_make_resource (r->start, ((r->end - r->start) + 1));
					if (!res) {
						goto no_memory;
					}

					dbg ("RESOURCE[%d] %08x - %08x (IO)\n", i, (u32)r->start, (u32)r->end);

					res->next = func->io_head;
					func->io_head = res;
				}
				/* Memory */
				else if (r->flags & IORESOURCE_MEM) {
					/* prefetch memory */
					if (r->flags & IORESOURCE_PREFETCH) {
						res = acpiphp_make_resource (r->start, ((r->end - r->start) + 1));
						if (!res) {
							goto no_memory;
						}

						dbg ("RESOURCE[%d] %08x - %08x (PMEM)\n", i, (u32)r->start, (u32)r->end);

						res->next = func->p_mem_head;
						func->p_mem_head = res;
					}
					/* regular memory */
					else {
						res = acpiphp_make_resource (r->start, ((r->end - r->start) + 1));
						if (!res) {
							goto no_memory;
						}

						dbg ("RESOURCE[%d] %08x - %08x (MEM)\n", i, (u32)r->start, (u32)r->end);

						res->next = func->mem_head;
						func->mem_head = res;
					}
				}
			}

			/* check bus resource */
			bus = dev->subordinate;
			res = acpiphp_make_resource ((u64)bus->number, (u32)bus->subordinate - bus->number + 1);
			if (!res) {
				goto no_memory;
			}

			dbg ("RESOURCE %02x - %02x (BUS)\n", (u32)bus->number, (u32)bus->subordinate);

			res->next = func->bus_head;
			func->bus_head = res;
		}
	}

#if 1
	acpiphp_dump_func_resource(func);
#endif

	return 0;

 no_memory:
	err("out of memory\n");
	acpiphp_free_resource(&func->io_head);
	acpiphp_free_resource(&func->mem_head);
	acpiphp_free_resource(&func->p_mem_head);

	return -1;
}


/**
 * acpiphp_configure_slot - allocate PCI resources
 * @slot: slot to be configured
 *
 * initializes a PCI functions on a device inserted
 * into the slot
 *
 */
int acpiphp_configure_slot (struct acpiphp_slot *slot)
{
	struct acpiphp_func *func;
	struct list_head *l;
	u8 hdr;
	u32 dvid;
	int retval = 0;
	int is_multi = 0;

	pci_bus_read_config_byte(slot->bridge->pci_bus,
				 PCI_DEVFN(slot->device, 0),
				 PCI_HEADER_TYPE, &hdr);

	if (hdr & 0x80)
		is_multi = 1;

	list_for_each (l, &slot->funcs) {
		func = list_entry(l, struct acpiphp_func, sibling);
		if (is_multi || func->function == 0) {
			pci_bus_read_config_dword(slot->bridge->pci_bus,
						  PCI_DEVFN(slot->device,
							    func->function),
						  PCI_VENDOR_ID, &dvid);
			if (dvid != 0xffffffff) {
				retval = init_config_space(func);
				if (retval)
					break;
			}
		}
	}

	return retval;
}

/**
 * acpiphp_configure_function - configure PCI function
 * @func: function to be configured
 *
 * initializes a PCI functions on a device inserted
 * into the slot
 *
 */
int acpiphp_configure_function (struct acpiphp_func *func)
{
	/* all handled by the pci core now */
	return 0;
}

/**
 * acpiphp_unconfigure_function - unconfigure PCI function
 * @func: function to be unconfigured
 *
 */
void acpiphp_unconfigure_function (struct acpiphp_func *func)
{
	struct acpiphp_bridge *bridge;

	/* if pci_dev is NULL, ignore it */
	if (!func->pci_dev)
		return;

	pci_remove_bus_device(func->pci_dev);

	/* free all resources */
	bridge = func->slot->bridge;

	spin_lock(&bridge->res_lock);
	acpiphp_move_resource(&func->io_head, &bridge->io_head);
	acpiphp_move_resource(&func->mem_head, &bridge->mem_head);
	acpiphp_move_resource(&func->p_mem_head, &bridge->p_mem_head);
	acpiphp_move_resource(&func->bus_head, &bridge->bus_head);
	spin_unlock(&bridge->res_lock);
}

/*
 * For P2P bridge hotplug support.
 * check and initialize configuration register of
 *                                   behind P2P bridge device.
 */
static int
init_config_space_behind_bridge (struct acpiphp_func *func)
{
	struct acpiphp_bridge	*bridge;
	struct pci_bus			*bus;
	struct pci_resource		*res;
	int tmp;
	int devfn;
	u8	io_base;
	u16	pfmem_base;
	u8	bus_no;

	/* initialize member data */
	bridge	= func->slot->bridge;
	bus		= bridge->pci_bus;
	res		= NULL;
	bus_no	= bus->number;
	devfn	= PCI_DEVFN (func->slot->device, func->function);

	dbg("Configure bridge device [%02x:%02x.%d]\n",
						bus_no, func->slot->device, func->function);

	/* set PCI bus number for PCI configuration access */
	acpiphp_pci_bus->number = bus_no;

	/* set primary bus number */
	dbg("---> Primary     : 0x%02x\n", bus_no);
	pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
								PCI_PRIMARY_BUS, bus_no);

	/* check bus resource for secondary bus */
	res = find_secondary_busno (func);
	if (!res) {
		return -1;
	}
	res->next = func->bus_head;
	func->bus_head = res;

	/* set secondary bus number */
	dbg("---> Secondary   : 0x%02x\n", (u8)res->base);
	pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
								PCI_SECONDARY_BUS, (u8)res->base);

	/* set subordinate bus number */
	dbg("---> Subordinate : 0x%02x\n", (u8)res->base);
	pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
								PCI_SUBORDINATE_BUS, (u8)res->base);

	/* set cache line size */
	dbg("---> Cache line  : 0x%02x\n", ACPIPHP_CACHE_LINE);
	pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
								PCI_CACHE_LINE_SIZE, ACPIPHP_CACHE_LINE);

	/* set latency timer */
	dbg("---> Latency     : 0x%02x\n", ACPIPHP_LATENCY);
	pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
								PCI_LATENCY_TIMER, ACPIPHP_LATENCY);

	/* set secondary latency timer */
	dbg("---> Sec Latency : 0x%02x\n", ACPIPHP_SEC_LATENCY);
	pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
								PCI_SEC_LATENCY_TIMER, ACPIPHP_SEC_LATENCY);

	/********************************************************/
	/*  scan device under PCI-to-PCI bridge                 */
	/********************************************************/
	tmp = scan_behind_bridge (func);
	if (tmp < 0) {
		acpiphp_move_resource (&func->bus_head, &bridge->bus_head);
		return -1;	/* error */
	}

	/* get PCI_IO/PREF_RANGE_TYPE */
	acpiphp_pci_bus->number = bus_no;
	pci_bus_read_config_byte (acpiphp_pci_bus, devfn,
								PCI_IO_BASE, &io_base);
	pci_bus_read_config_word (acpiphp_pci_bus, devfn,
								PCI_PREF_MEMORY_BASE, &pfmem_base);

	/* set I/O base and limit */
	if (func->io_head) {
		dbg("---> I/O         : base 0x%02x limit 0x%02x\n", 
										(u8)(0xf0 & (func->io_head->base >> 8)),
										(u8)(0xf0 & ((func->io_head->base + func->io_head->length -1) >> 8)));
		pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
									PCI_IO_BASE, 
									(u8)(0xf0 & (func->io_head->base >> 8)));
		pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
									PCI_IO_LIMIT,
									(u8)(0xf0 & ((func->io_head->base + func->io_head->length - 1) >> 8)));
		if ((io_base & PCI_IO_RANGE_TYPE_MASK) == PCI_IO_RANGE_TYPE_32) {
			dbg("---> I/O Type32  : base 0x0000 limit 0x0000\n");
			pci_bus_write_config_word (acpiphp_pci_bus, devfn,
										PCI_IO_BASE_UPPER16, 0x0000);
			pci_bus_write_config_word (acpiphp_pci_bus, devfn,
										PCI_IO_LIMIT_UPPER16, 0x0000);
		}
	}
	else {
		dbg("---> I/O         : base 0x%02x limit 0x%02x\n", 0x00, 0x00);
		pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
									PCI_IO_BASE, 0xff);
		pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
									PCI_IO_LIMIT, 0x00);
	}

	/* set Memory base and limit */
	if (func->mem_head) {
		dbg("---> Memory      : base 0x%04x limit 0x%04x\n",
									(u16)(0xfff0 & (func->mem_head->base >> 16)),
									(u16)(0xfff0 & ((func->mem_head->base + func->mem_head->length - 1) >> 16)));
		pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_MEMORY_BASE,
									(u16)(0xfff0 & (func->mem_head->base >> 16)));
		pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_MEMORY_LIMIT,
									(u16)(0xfff0 & ((func->mem_head->base + func->mem_head->length - 1) >> 16)));
	}
	else {
		dbg("---> Memory      : base 0xffff limit 0x0000\n");
		pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_MEMORY_BASE, 0xffff);
		pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_MEMORY_LIMIT, 0x0000);
	}

	/* set Prefetch memory base and limit */
	if (func->p_mem_head) {
		dbg("---> Pre Memory  : base 0x%04x limit 0x%04x\n",
									(u16)(0xfff0 & (func->p_mem_head->base >> 16)),
									(u16)(0xfff0 & ((func->p_mem_head->base + func->p_mem_head->length - 1) >> 16)));
		pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_PREF_MEMORY_BASE,
									(u16)(0xfff0 & (func->p_mem_head->base >> 16)));
		pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_PREF_MEMORY_LIMIT,
									(u16)(0xfff0 & ((func->p_mem_head->base + func->p_mem_head->length - 1) >> 16)));
		if ((pfmem_base & PCI_PREF_RANGE_TYPE_MASK) == PCI_PREF_RANGE_TYPE_64) {
			dbg("---> MEM Type64  : base 0x00000000 limit:0x00000000\n");
			pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
										PCI_PREF_BASE_UPPER32, 0x00000000);
			pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
										PCI_PREF_LIMIT_UPPER32, 0x00000000);
		}
	}
	else {
		dbg("---> Pre Memory  : base 0xffff limit 0x0000\n");
		pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_PREF_MEMORY_BASE, 0xffff);
		pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_PREF_MEMORY_LIMIT, 0x0000);
	}

	/* set Command register */
	dbg("---> Command     : 0x%04x\n", ACPIPHP_DEVICEENABLE);
	pci_bus_write_config_word (acpiphp_pci_bus, devfn,
								PCI_COMMAND, ACPIPHP_DEVICEENABLE);

	/* set Bridge control */
	dbg("---> Bridge ctrl : 0x%04x\n", ACPIPHP_BRIDGE_CTRL);
	pci_bus_write_config_word (acpiphp_pci_bus, devfn,
								PCI_BRIDGE_CONTROL, ACPIPHP_BRIDGE_CTRL);

	return 0;
}

/*
 * find scondary bus number
 */
static struct pci_resource *
find_secondary_busno (struct acpiphp_func *func)
{
	struct acpiphp_bridge	*bridge;
	struct pci_bus			*bus;
	struct pci_resource		*res;

	bridge	= func->slot->bridge;
	bus		= bridge->pci_bus;
	res		= NULL;

	/* get usable resource range */
	spin_lock (&bridge->res_lock);
	res = acpiphp_get_resource_with_base (&bridge->bus_head,
											(u64)bridge->bus_head->base,
											(u32)1);
	spin_unlock (&bridge->res_lock);

	if (!res) {
		err ("cannot allocate requested bus for %02x:%02x.%d\n",
				bus->number,
				func->slot->device,
				func->function);
		return NULL;
	}

	return res;
}

/*
 * scan behind bridge
 */
static int
scan_behind_bridge (struct acpiphp_func *func)
{
	struct acpiphp_bridge	*bridge;
	struct acpiphp_func		*usable_res;
	struct acpiphp_func		*used_res;
	u8 device;
	u8 function;
	u8 hdrtype;
	u8 bus_no;
	u16 vendor_id;
	u32 class;
	int is_multi;
	int tmp;
	int devfn;

	/* initialize */
	bridge		= func->slot->bridge;
	bus_no		= func->bus_head->base;
	is_multi	= 0;

	/* initialize resource struct */
	usable_res = kmalloc(sizeof(struct acpiphp_func), GFP_KERNEL);
	if (!usable_res) {
		err ("cannot create usable resource struct\n");
		return -1;
	}
	memset (usable_res, 0, sizeof(struct acpiphp_func));

	used_res = kmalloc(sizeof(struct acpiphp_func), GFP_KERNEL);
	if (!used_res) {
		err ("cannot create usable resource struct\n");
		kfree(usable_res);
		return -1;
	}
	memset (used_res, 0, sizeof(struct acpiphp_func));


	/* get usable resource */
	spin_lock (&bridge->res_lock);
	usable_res->io_head		= acpiphp_get_io_resource (&bridge->io_head, 0x1000);
	usable_res->mem_head	= acpiphp_get_resource (&bridge->mem_head, 0x100000);
	usable_res->p_mem_head	= acpiphp_get_resource (&bridge->p_mem_head, 0x100000);
	spin_unlock (&bridge->res_lock);


	/* check all slots */
	for (device = 0; device < 32; device++) {

		/* check all functions */
		for (function = 0; function < 8; function++) {

			/* check multi function */
			if (function && !is_multi) {
				continue;
			}

			/* initialize scanning bus */
			acpiphp_pci_bus->number = bus_no;

			/* check enable device */
			devfn = PCI_DEVFN(device, function);
			pci_bus_read_config_word (acpiphp_pci_bus, devfn,
										PCI_VENDOR_ID, &vendor_id);
			if (vendor_id != 0xFFFF) {	/* device is exist */

				dbg("When bus scaned, new device found! [%02x:%02x.%d]\n", bus_no, device, function);

				pci_bus_read_config_byte (acpiphp_pci_bus, devfn,
											PCI_HEADER_TYPE, &hdrtype);
				pci_bus_read_config_dword (acpiphp_pci_bus, devfn,
											PCI_CLASS_REVISION, &class);
				if (!function) {
					is_multi = hdrtype & 0x80;
				}

				/* check enable device type */
				if (hdrtype & PCI_HEADER_TYPE_BRIDGE) {
					err("embedded bridges not supported for hot-plugging.(b:%x d:%x f:%d)\n",
													bus_no, device, function);
					continue;		/* goto next function */
				}

				class >>= 8;	/* to take revision ID */
				if (class == PCI_CLASS_NOT_DEFINED_VGA) {
					err("The device is VGA compatible and as is not supported for hot-plugging.(b:%x d:%x f:%d)\n",
													bus_no, device, function);
					continue;		/* goto next function */
				}
				else if (class == PCI_CLASS_DISPLAY_VGA) {
					err("The device is not supported for hot-plugging.(b:%x d:%x f:%d)\n",
													bus_no, device, function);
					continue;		/* goto next function */
				}

				/* initialize base address register behind bridge */
				tmp = init_base_address_registers (	func,
													device,
													function,
													usable_res,
													used_res);
				if (tmp < 0) {
					err("Initialize base address register error.[%02x:%02x.%d]\n", bus_no, device, function);
					/* return used resource */
					acpiphp_move_resource (&used_res->io_head, &usable_res->io_head);
					acpiphp_move_resource (&used_res->mem_head, &usable_res->mem_head);
					acpiphp_move_resource (&used_res->p_mem_head, &usable_res->p_mem_head);

					continue;		/* goto next function */
				}

				/* set used resource to func */
				acpiphp_move_resource (&used_res->io_head, &func->io_head);
				acpiphp_move_resource (&used_res->mem_head, &func->mem_head);
				acpiphp_move_resource (&used_res->p_mem_head, &func->p_mem_head);

			}	/* end device config */

		}	/* end check function */

	}	/* end device function */


	/* check used resources and move rest resources */
	/* I/O */
	if (func->io_head) {
		/* move all resource from usable to func */
		acpiphp_move_resource (&usable_res->io_head, &func->io_head);
		acpiphp_resource_sort_and_combine (&func->io_head);
	}
	else {
		/* return resource from usable to bridge */
			spin_lock (&bridge->res_lock);
			acpiphp_move_resource (&usable_res->io_head, &bridge->io_head);
			spin_unlock (&bridge->res_lock);
			acpiphp_resource_sort_and_combine (&bridge->io_head);
	}
	/* Memory */
	if (func->mem_head) {
		/* move all resource from usable to func */
		acpiphp_move_resource (&usable_res->mem_head, &func->mem_head);
		acpiphp_resource_sort_and_combine (&func->mem_head);
	}
	else {
		/* return resource from usable to bridge */
		spin_lock (&bridge->res_lock);
		acpiphp_move_resource (&usable_res->mem_head, &bridge->mem_head);
		spin_unlock (&bridge->res_lock);
		acpiphp_resource_sort_and_combine (&bridge->mem_head);
	}
	/* Prefetch memory */
	if (func->p_mem_head) {
		/* move all resource from usable to func */
		acpiphp_move_resource (&usable_res->p_mem_head, &func->p_mem_head);
		acpiphp_resource_sort_and_combine (&func->p_mem_head);
	}
	else {
		/* return resource from usable to bridge */
		spin_lock (&bridge->res_lock);
		acpiphp_move_resource (&usable_res->p_mem_head, &bridge->p_mem_head);
		spin_unlock (&bridge->res_lock);
		acpiphp_resource_sort_and_combine (&bridge->p_mem_head);
	}

	/* delete resource area */
	kfree (usable_res);
	kfree (used_res);

	return 0;
}

/*
 * Initialize base address register 
 */
static int
init_base_address_registers (struct acpiphp_func *func, u8 devno, u8 funcno, struct acpiphp_func *usable_res, struct acpiphp_func *used_res)
{
	u32 bar;
	u32 len;
	u32 address[] = {
		PCI_BASE_ADDRESS_0,
		PCI_BASE_ADDRESS_1,
		PCI_BASE_ADDRESS_2,
		PCI_BASE_ADDRESS_3,
		PCI_BASE_ADDRESS_4,
		PCI_BASE_ADDRESS_5,
		0
	};
	int count;
	struct acpiphp_bridge	*bridge;
	struct pci_resource		*res;
	u8 busno;
	unsigned int devfn;
	u16 tmp;

	/* initialize */
	bridge	= func->slot->bridge;
	busno	= func->bus_head->base;
	devfn	= PCI_DEVFN(devno, funcno);

	/* set PCI bus number for PCI configuration access */
	acpiphp_pci_bus->number = busno;

	/* check base address register */
	for (count = 0; address[count]; count++) {

		pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
										address[count], 0xFFFFFFFF);
		pci_bus_read_config_dword (acpiphp_pci_bus, devfn,
										address[count], &bar);

		if (!bar) {		/* This BAR is not implemented */
			continue;
		}

		/* This is I/O */
		if (bar & PCI_BASE_ADDRESS_SPACE_IO) {

			len = bar & (PCI_BASE_ADDRESS_IO_MASK & 0xFFFF);
			len = len & ~(len - 1);

			res = acpiphp_get_io_resource (&usable_res->io_head, len);
			if (!res) {
				err ("cannot allocate requested io for %02x:%02x.%d len %x BAR %d\n",
						busno, devno, funcno, len, count);
				return -1;
			}

			dbg ("len in I/O %x, BAR %d. res 0x%08x - 0x%08x\n", 
						len, count, (u32)res->base, (u32)(res->base + res->length - 1));

			pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
											address[count], (u32)res->base);
			res->next			= used_res->io_head;
			used_res->io_head	= res;
		}
		/* This is Memory */
		else {
			/* This is Prefetch Memory */
			if (bar & PCI_BASE_ADDRESS_MEM_PREFETCH) {

				len = bar & 0xFFFFFFF0;
				len = ~len + 1;

				res = acpiphp_get_resource (&usable_res->p_mem_head, len);
				if (!res) {
					err ("cannot allocate requested pfmem for %02x:%02x.%d len %x BAR %d\n",
							busno, devno, funcno, len, count);
					return -1;
				}

				dbg ("len in PFMEM %x, BAR %d. res 0x%08x - 0x%08x\n",
						len, count, (u32)res->base, (u32)(res->base + res->length - 1));

				pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
												address[count], (u32)res->base);
				if (bar & PCI_BASE_ADDRESS_MEM_TYPE_64) {
					count += 1;
					pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
													address[count], (u32)(res->base >> 32));
				}
				res->next				= used_res->p_mem_head;
				used_res->p_mem_head	= res;
			}
			/* This is regular memory */
			else {

				len = bar & 0xFFFFFFF0;
				len = ~len + 1;

				res = acpiphp_get_resource (&usable_res->mem_head, len);
				if (!res) {
					err ("cannot allocate requested mem for %02x:%02x.%d len %x BAR %d\n",
							busno, devno, funcno, len, count);
					return -1;
				}

				dbg ("len in MEM %x, BAR %d. res 0x%08x - 0x%08x\n",
						len, count, (u32)res->base, (u32)(res->base + res->length - 1));

				pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
												address[count], (u32)res->base);
				if (bar & PCI_BASE_ADDRESS_MEM_TYPE_64) {
					count += 1;
					pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
													address[count], (u32)(res->base >> 32));
				}
				res->next			= used_res->mem_head;
				used_res->mem_head	= res;
			}
		}
	}

	/* disable expansion rom */
	pci_bus_write_config_dword (acpiphp_pci_bus, devfn,
									PCI_ROM_ADDRESS, 0x00000000);

	/* set PCI parameters from _HPP */
	pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
									PCI_CACHE_LINE_SIZE, bridge->hpp.cache_line_size);
	pci_bus_write_config_byte (acpiphp_pci_bus, devfn,
									PCI_LATENCY_TIMER, bridge->hpp.latency_timer);

	pci_bus_read_config_word (acpiphp_pci_bus, devfn,
									PCI_COMMAND, &tmp);
	if (bridge->hpp.enable_SERR)
		tmp |= PCI_COMMAND_SERR;
	if (bridge->hpp.enable_PERR)
		tmp |= PCI_COMMAND_PARITY;
	pci_bus_write_config_word (acpiphp_pci_bus, devfn,
									PCI_COMMAND, tmp);

	dbg ("set PCI param. cache_line:0x%02x latency:0x%02x command:0x%04x\n",
				bridge->hpp.cache_line_size, bridge->hpp.latency_timer, tmp);

	return 0;	/* success */
}

