/*
 * QLogic Fibre Channel HBA Driver
 * Copyright (c)  2003-2005 QLogic Corporation
 *
 * See LICENSE.qla2xxx for copyright and licensing details.
 */

/*
 * QLogic ISP2x00 Multi-path LUN Support Driver 
 * Linux specific functions
 *
 */

#include "qla_def.h"
#include "qla_foln.h"

#define MAX_SEARCH_STR_SIZE	512

/*
 * qla2x00_set_lun_data_from_config
 * Set lun_data byte from the configuration parameters.
 *
 * Input:
 * host -- pointer to host adapter structure.
 * port -- pointer to port
 * tgt  -- target number
 * dev_no  -- device number
 */
void
qla2x00_set_lun_data_from_config(mp_host_t *host, fc_port_t *port,
    uint16_t tgt, uint16_t dev_no)
{
	char		*propbuf;  /* As big as largest search string */
	int		rval;
	int16_t		lun, l;
	scsi_qla_host_t *ha = host->ha;
	mp_device_t	*dp;
	lun_bit_mask_t	*plun_mask;
	lun_bit_mask_t  *mask_ptr;
	mp_path_list_t	*pathlist;

	mp_path_t *path;

	if (qla2x00_get_ioctl_scrap_mem(ha, (void **)&propbuf,
	    MAX_SEARCH_STR_SIZE)) {
		/* not enough memory */
		DEBUG9_10(printk("%s(%ld): inst=%ld scrap not big enough. "
		    "propbuf requested=%d.\n",
		    __func__, ha->host_no, ha->instance,
		    MAX_SEARCH_STR_SIZE));
		return;
	}

	if (qla2x00_get_ioctl_scrap_mem(ha, (void **)&plun_mask,
	    sizeof(lun_bit_mask_t))) {
		/* not enough memory */
		DEBUG9_10(printk("%s(%ld): inst=%ld scrap not big enough. "
		    "lun_mask requested=%ld.\n",
		    __func__, ha->host_no, ha->instance,
		    (ulong)sizeof(lun_bit_mask_t)));
		qla2x00_free_ioctl_scrap_mem(ha);
		return;
	}
	mask_ptr = plun_mask;

	dp = host->mp_devs[tgt];
	if (dp == NULL) {
		printk("qla2x00_set_lun_data_from_config: Target %d "
		    "not found for hba %d\n",tgt, host->instance);
		qla2x00_free_ioctl_scrap_mem(ha);
		return;
	}
	if ( (pathlist = dp->path_list) == NULL ) {
		printk("qla2x00_set_lun_data_from_config: path list "
		    "not found for target %d\n", tgt);
		qla2x00_free_ioctl_scrap_mem(ha);
		return;
	}

	if ((path = qla2x00_find_path_by_name(host, pathlist,
	    port->port_name)) == NULL ) {
		printk("qla2x00_set_lun_data_from_config: No path found "
		    "for target %d\n", tgt);
		qla2x00_free_ioctl_scrap_mem(ha);
		return;
	}

	/* Get "target-N-device-N-preferred" as a 256 bit lun_mask*/
	sprintf(propbuf, "scsi-qla%ld-tgt-%d-di-%d-preferred", ha->instance,
	    tgt, dev_no);
	DEBUG3(printk("build_tree: %s\n",propbuf));

	rval = qla2x00_get_prop_xstr(ha, propbuf,
	    (uint8_t *)(plun_mask), sizeof(lun_bit_mask_t));

	if (rval == -1) {
		/* EMPTY */
		DEBUG2(printk("%s(%ld): no preferred mask entry found for "
		    "path id %d on port %02x%02x%02x%02x%02x%02x%02x%02x.\n",
		    __func__, ha->host_no, path->id,
		    path->portname[0], path->portname[1],
		    path->portname[2], path->portname[3],
		    path->portname[4], path->portname[5],
		    path->portname[6], path->portname[7]));
	} else {
		if (rval != sizeof(lun_bit_mask_t)) {
			/* EMPTY */
			printk("qla2x00_set_lun_data_from_config: "
			    "Preferred mask len %d is incorrect.\n", rval);
		}

		DEBUG3(printk("%s(%ld): reading Preferred Mask for path id %d "
		    "on port %02x%02x%02x%02x%02x%02x%02x%02x:\n",
		    __func__, ha->host_no, path->id,
		    path->portname[0], path->portname[1],
		    path->portname[2], path->portname[3],
		    path->portname[4], path->portname[5],
		    path->portname[6], path->portname[7]));
		DEBUG3(qla2x00_dump_buffer((char *)plun_mask,
		    sizeof(lun_bit_mask_t)));

		for (lun = MAX_LUNS-1, l =0; lun >= 0; lun--, l++ ) {
			if (EXT_IS_LUN_BIT_SET(mask_ptr, lun)) {
				path->lun_data.data[l] |=
				    LUN_DATA_PREFERRED_PATH;
				pathlist->current_path[l] = path->id;
			} else {
				path->lun_data.data[l] &=
				    ~LUN_DATA_PREFERRED_PATH;
			}
		}

	}

	/* Get "target-N-device-N-lun-disable" as a 256 bit lun_mask*/
	sprintf(propbuf, "scsi-qla%ld-tgt-%d-di-%d-lun-disabled", ha->instance,
	    tgt, dev_no);
	DEBUG3(printk("build_tree: %s\n",propbuf));

	rval = qla2x00_get_prop_xstr(ha, propbuf,
	    (uint8_t *)plun_mask, sizeof(lun_bit_mask_t));
	if (rval == -1) {
		/* default: all luns enabled */
		DEBUG3(printk("%s(%ld): no entry found for path id %d. "
		    "Assume all LUNs enabled on port %02x%02x%02x%02x%02x%"
		    "02x%02x%02x.\n",
		    __func__, ha->host_no, path->id,
		    path->portname[0], path->portname[1],
		    path->portname[2], path->portname[3],
		    path->portname[4], path->portname[5],
		    path->portname[6], path->portname[7]));

		for (lun = 0; lun < MAX_LUNS; lun++) {
			path->lun_data.data[lun] |= LUN_DATA_ENABLED;
		}
	} else {
		if (rval != sizeof(lun_bit_mask_t)) {
			printk("qla2x00_set_lun_data_from_config: Enable "
			    "mask has wrong size %d != %ld\n",
			    rval, (ulong)sizeof(lun_bit_mask_t));
		} else {
			for (lun = MAX_LUNS-1, l =0; lun >= 0; lun--, l++) {
				/* our bit mask is inverted */
				if (!EXT_IS_LUN_BIT_SET(mask_ptr,lun))
					path->lun_data.data[l] |=
					    LUN_DATA_ENABLED;
				else
					path->lun_data.data[l] &=
					    ~LUN_DATA_ENABLED;
			}
			DEBUG3(printk("%s(%ld): got lun mask for path id %d "
			    "port %02x%02x%02x%02x%02x%02x%02x%02x:\n",
			    __func__, ha->host_no, path->id,
			    path->portname[0], path->portname[1],
			    path->portname[2], path->portname[3],
			    path->portname[4], path->portname[5],
			    path->portname[6], path->portname[7]));
			DEBUG3(qla2x00_dump_buffer(
			    (uint8_t *)&path->lun_data.data[0], 64));
		}
	}

	DEBUG3(printk("qla2x00_set_lun_data_from_config: Luns data for "
	    "device %p, instance %d, path id=%d\n",
	    dp,host->instance,path->id));
	DEBUG3(qla2x00_dump_buffer((char *)&path->lun_data.data[0], 64));

	qla2x00_free_ioctl_scrap_mem(ha);
	LEAVE("qla2x00_set_lun_data_from_config");
}



/*
 * qla2x00_cfg_build_path_tree
 *	Find all path properties and build a path tree. The
 *  resulting tree has no actual port assigned to it
 *  until the port discovery is done by the lower level.
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Context:
 *	Kernel context.
 */
void
qla2x00_cfg_build_path_tree(scsi_qla_host_t *ha)
{
	char		*propbuf;
	uint8_t		node_name[WWN_SIZE];
	uint8_t		port_name[WWN_SIZE];
	fc_port_t	*port;
	uint16_t	dev_no = 0, tgt;
	int		instance, rval;
	mp_host_t	*host = NULL;
	int		done;
	uint8_t         control_byte;


	ENTER("qla2x00_cfg_build_path_tree");

	printk(KERN_INFO
	    "qla02%d: ConfigRequired is set. \n", (int)ha->instance);
	DEBUG(printk("qla2x00_cfg_build_path_tree: hba =%ld\n",
	    ha->instance));

	if (qla2x00_get_ioctl_scrap_mem(ha, (void **)&propbuf,
	    MAX_SEARCH_STR_SIZE)) {
		/* not enough memory */
		DEBUG9_10(printk("%s(%ld): inst=%ld scrap not big enough. "
		    "propbuf requested=%d.\n",
		    __func__, ha->host_no, ha->instance,
		    MAX_SEARCH_STR_SIZE));
		return;
	}

	/* Look for adapter nodename in properties */
	sprintf(propbuf, "scsi-qla%ld-adapter-port", ha->instance);
	DEBUG(printk("build_tree: %s\n",propbuf));

	rval = qla2x00_get_prop_xstr(ha, propbuf, port_name, WWN_SIZE);
	if (rval != WWN_SIZE) {
		qla2x00_free_ioctl_scrap_mem(ha);
		return;
	}

	/* Does nodename match the host adapter nodename? */
	if (!qla2x00_is_nodename_equal(ha->port_name, port_name)) {
		printk(KERN_INFO
		    "scsi(%d): Adapter nodenames don't match - ha = %p.\n",
		    (int)ha->instance,ha);
		DEBUG(printk("qla(%d): Adapter nodenames don't match - "
		    "ha=%p. port name=%02x%02x%02x%02x%02x%02x%02x%02x\n",
		    (int)ha->instance,ha, ha->port_name[0], ha->port_name[1],
		    ha->port_name[2], ha->port_name[3], ha->port_name[4],
		    ha->port_name[5], ha->port_name[6], ha->port_name[7]));

		qla2x00_free_ioctl_scrap_mem(ha);
		return;
	}

	DEBUG(printk("%s: found entry for adapter port %02x%02x%02x%02x"
	    "%02x%02x%02x%02x.\n",
	    __func__,
	    port_name[0], port_name[1], port_name[2],
	    port_name[3], port_name[4], port_name[5],
	    port_name[6], port_name[7]));

	instance = ha->instance;
	if ((host = qla2x00_alloc_host(ha)) == NULL) {
		printk(KERN_INFO
		    "scsi(%d): Couldn't allocate host - ha = %p.\n",
		    (int)instance,ha);
	} else {
		/* create a dummy port */
		port = kmalloc(sizeof(fc_port_t), GFP_KERNEL);
		if (port == NULL) {
			printk(KERN_INFO
			    "scsi(%d): Couldn't allocate port.\n",
			    (int)instance);
			DEBUG(printk("qla(%d): Couldn't allocate port.\n",
			    (int)host->instance));
			/* remove host */
			qla2x00_free_ioctl_scrap_mem(ha);
			return;
		}
		/* Setup fcport template structure. */
		memset(port, 0, sizeof (fc_port_t));
		port->port_type = FCT_UNKNOWN;
		port->loop_id = FC_NO_LOOP_ID;
		port->iodesc_idx_sent = IODESC_INVALID_INDEX;
		atomic_set(&port->state, FCS_UNCONFIGURED);
		port->flags = FC_RLC_SUPPORT;
		INIT_LIST_HEAD(&port->fcluns);

		done = 0;

		/* For each target on the host bus adapter */
		for (tgt = 0; tgt < MAX_MP_DEVICES &&
		    !done; tgt++) {

			/* get all paths for this target */
			for (dev_no = 0; dev_no < MAX_PATHS_PER_DEVICE &&
			    !done ; dev_no++) {

				/*
				 * O(N*M) scan, should ideally check if there
				 * are any tgt entries present, if not, then
				 * continue.
				 *
				 *   sprintf(propbuf,
				 * 		"scsi-qla%d-tgt-%d-",
				 *		instance, tgt);
				 *   if (strstr(ha->cmdline, propbuf) == NULL)
				 *	continue;
				 *
				 */
				memset(port, 0, sizeof (fc_port_t));

				/*
				 * Get "target-N-device-N-node" is a 16-chars
				 * number
				 */
				sprintf(propbuf,
				    "scsi-qla%ld-tgt-%d-di-%d-node",
				    ha->instance, tgt, dev_no);

				rval = qla2x00_get_prop_xstr(ha, propbuf,
				    node_name, WWN_SIZE);
				if (rval != WWN_SIZE)
					/* di values may not be contiguous for
					 * override case.
					 */
					continue;

				DEBUG(printk("build_tree: %s\n",propbuf));
				memcpy(port->node_name, node_name, WWN_SIZE);

				/*
				 * Get "target-N-device-N-port" is a 16-chars
				 * number
				 */
				sprintf(propbuf,
				    "scsi-qla%ld-tgt-%d-di-%d-port",
				    ha->instance, tgt, dev_no);

				rval = qla2x00_get_prop_xstr(ha, propbuf,
				    port_name, WWN_SIZE);
				if (rval != WWN_SIZE)
					continue;

				DEBUG(printk("build_tree: %s\n",propbuf));
				memcpy(port->node_name, node_name, WWN_SIZE);
				memcpy(port->port_name, port_name, WWN_SIZE);
				port->flags |= FC_CONFIG;

				/*
				 * Get "target-N-device-N-control" if property 
				 * is present then all luns are visible.
				 */
				sprintf(propbuf,
				    "scsi-qla%ld-tgt-%d-di-%d-control",
				    ha->instance, tgt, dev_no);
				rval = qla2x00_get_prop_xstr(ha, propbuf,
				    (uint8_t *)(&control_byte),
				    sizeof(control_byte));
				if (rval == -1) {
					/* error getting string. go to next. */
					DEBUG2(printk(
					    "%s: string parsing failed.\n",
					    __func__));
					continue;
				}

				DEBUG3(printk("build_tree: %s\n",propbuf));

				DEBUG(printk("build_tree: control byte 0x%x\n",
				    control_byte));

				port->mp_byte = control_byte;
				DEBUG(printk("%s(%ld): calling update_mp_device"
				    " for host %p port %p-%02x%02x%02x%02x%02x"
				    "%02x%02x%02x tgt=%d mpbyte=%02x.\n",
				    __func__, ha->host_no, host, port,
				    port->port_name[0], port->port_name[1],
				    port->port_name[2], port->port_name[3],
				    port->port_name[4], port->port_name[5],
				    port->port_name[6], port->port_name[7],
				    tgt, port->mp_byte));

				qla2x00_update_mp_device(host, port, tgt,
				    dev_no);

				/* free any mplun info */

				qla2x00_set_lun_data_from_config(host,
				    port, tgt, dev_no);
			}
		}
		kfree(port);
	}

	qla2x00_free_ioctl_scrap_mem(ha);

	LEAVE("qla2x00_cfg_build_path_tree");
	DEBUG(printk("Leaving: qla2x00_cfg_build_path_tree\n"));
}

/*
 * qla2x00_cfg_display_devices
 *      This routine will the node names of the different devices found
 *      after port inquiry.
 *
 * Input:
 *
 * Returns:
 *      None.
 */
void
qla2x00_cfg_display_devices(int flag)
{
	mp_host_t     *host;
	int     id;
	mp_device_t	*dp;
	mp_path_t  *path;
	mp_path_list_t	*path_list;
	int cnt, i, dev_no;
	int instance;
	lun_bit_mask_t	lun_mask;
	int	mask_set;
	uint8_t	l;
	mp_lun_t	*lun;
	unsigned char 	tmp_buf[32];

	for (host = mp_hosts_base; (host); host = host->next) {

		instance = (int) host->instance;
		/* Display the node name for adapter */
		printk(KERN_INFO
			"scsi-qla%d-adapter-port="
			"%02x%02x%02x%02x%02x%02x%02x%02x\\;\n",
			instance,
			host->portname[0],
			host->portname[1],
			host->portname[2],
			host->portname[3],
			host->portname[4],
			host->portname[5],
			host->portname[6],
			host->portname[7]);

		for (id = 0; id < MAX_MP_DEVICES; id++) {
			if( (dp = host->mp_devs[id] ) == NULL )
				continue;

			path_list = dp->path_list;


			if( (path = path_list->last) != NULL ) {
				/* Print out device port names */
				path = path->next; /* first path */
				for (dev_no = 0,  cnt = 0;
					cnt < path_list->path_cnt;
					path = path->next, cnt++) {

					/* skip others if not our host */
					if (host != path->host)
						continue;
					printk(KERN_INFO
						"scsi-qla%d-tgt-%d-di-%d-node="
						"%02x%02x%02x%02x"
						"%02x%02x%02x%02x\\;\n",
						instance, id, path->id,
						dp->nodename[0],
						dp->nodename[1],
						dp->nodename[2],
						dp->nodename[3],
						dp->nodename[4],
						dp->nodename[5],
						dp->nodename[6],
						dp->nodename[7]);

					/* port_name */
					printk(KERN_INFO
						"scsi-qla%d-tgt-%d-di-%d-port="
						"%02x%02x%02x%02x"
						"%02x%02x%02x%02x\\;\n",
						instance, id, path->id,
						path->portname[0],
						path->portname[1],
						path->portname[2],
						path->portname[3],
						path->portname[4],
						path->portname[5],
						path->portname[6],
						path->portname[7]);

					/* control byte */
					printk(KERN_INFO
						"scsi-qla%d-tgt-%d-di-%d-"
						"control=%02x\\;\n",
						instance, id, path->id,
						path->mp_byte);

					/*
					 * Build preferred bit mask for this
					 * path */
					memset(&lun_mask, 0, sizeof(lun_mask));
					mask_set = 0;
					for (i = 0; i < MAX_LUNS; i++) {
						l = (uint8_t)(i & 0xFF);
						if (path_list->current_path[l] == path->id ) {
							EXT_SET_LUN_BIT((&lun_mask),l);
							mask_set++;
						}
					}
					if (mask_set) {
						printk(KERN_INFO
							"scsi-qla%d-tgt-%d-di-%d-preferred=%08x%08x%08x%08x%08x%08x%08x%08x\\;\n",
							instance,  id, path->id,
							*((uint32_t *) &lun_mask.mask[28]),
							*((uint32_t *) &lun_mask.mask[24]),
							*((uint32_t *) &lun_mask.mask[20]),
							*((uint32_t *) &lun_mask.mask[16]),
							*((uint32_t *) &lun_mask.mask[12]),
							*((uint32_t *) &lun_mask.mask[8]),
							*((uint32_t *) &lun_mask.mask[4]),
							*((uint32_t *) &lun_mask.mask[0]) );
					}
					/*
					 * Build disable bit mask for this path
					 */
					mask_set = 0;
					for (i = 0; i < MAX_LUNS; i++) {
						l = (uint8_t)(i & 0xFF);
						if (!(path->lun_data.data[l] &
							LUN_DATA_ENABLED) ) {

							mask_set++;
						}
					}
					if (mask_set) {
						printk(KERN_INFO
							"scsi-qla%d-tgt-%d-di-%d-lun-disable=%08x%08x%08x%08x%08x%08x%08x%08x\\;\n",
							instance,  id, path->id,
							*((uint32_t *) &lun_mask.mask[28]),
							*((uint32_t *) &lun_mask.mask[24]),
							*((uint32_t *) &lun_mask.mask[20]),
							*((uint32_t *) &lun_mask.mask[16]),
							*((uint32_t *) &lun_mask.mask[12]),
							*((uint32_t *) &lun_mask.mask[8]),
							*((uint32_t *) &lun_mask.mask[4]),
							*((uint32_t *) &lun_mask.mask[0]) );
					}
					/* display lun wwuln */
					if( flag )
					for (lun = dp->luns; lun != NULL ; lun = lun->next) {
						printk(KERN_INFO
							"scsi-qla%d-tgt-%d-di-%d-lun-%d-lunid=",
							instance,  id, path->id, lun->number);
						for (i = 0 ; i < lun->siz ;
							       	i++) {
							sprintf(tmp_buf+i,
								"%02x", 
							  lun->wwuln[i]);
						}
						printk(KERN_INFO "%s:%02d;\n",
							tmp_buf,lun->siz); 
					}
					dev_no++;
				}

			}
		}
	}
}


